/*--------------------------------------------------------------------
 * $Id: cloud3d.c,v 1.1 2005/12/13 14:52:21 bernhard Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <float.h>
#include "uvspec.h"
#include "cloud3d.h"
#include "ipa.h"
#include "../src_c/wcloud3d.h"
#include "../src_c/mystic.h"


static void cp_1D_to_3D (profile3D *p3D, profile *p);
static int compare_levels_df (double *z1, int n1, float  *z2, int n2);


/***********************************************************************************/
/* Function:                                                              @62_30i@ */
/* Description:                                                                    */
/*  Setup a 3D atmosphere structure, to be used as input to MYSTIC.                */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i62_30@ */
/***********************************************************************************/

int setup_atmosphere3D (input_struct input, output_struct *output)
{
  int kc=0, status=0;
  char wcloud3d_filename[FILENAME_MAX];
  
  /* non Henyey-Greenstein if user-defined phase function */
  output->wc.threed.nonHGcld = (input.rte.mc.momcld!=MOMCLD_NONE || 
				strlen(input.rte.mc.filename[FN_MC_PCLD]) > 0 ? 1 : 0);

  /* don't need to read cloud properties if not a 3D calculation */
  /* 3D cloud file defined or no - need to re-think once 3D and  */
  /* IPA are homogenized ???                                     */
  if (strlen(input.rte.mc.filename[FN_MC_WCLOUD])==0 && input.rte.solver != MONTECARLO)
    return 0;
  
  if (*input.rte.mc.filename[FN_MC_WCLOUD])
    strcpy (wcloud3d_filename,  input.rte.mc.filename[FN_MC_WCLOUD]);
  else 
    strcpy (wcloud3d_filename,  "./wcloud3D.dat");
  
  if (!input.quiet)
    fprintf (stderr, " ... reading 3D cloud data from %s\n", wcloud3d_filename);
  
  
  fprintf (stderr, " ... reading 3D cloud\n");
  
  /* read 3D cloud data */
  status = read_3D_cloud (wcloud3d_filename,
			  &(output->wc.threed.Nx), 
			  &(output->wc.threed.Ny), 
			  &(output->wc.threed.nlyr), 
			  &(output->wc.threed.delX), &(output->wc.threed.delY),
			  &(output->wc.threed.zd), 
			  &(output->wc.threed.k3D),
			  &(output->wc.threed.threed),
			  &(output->wc.threed.g3D),
			  &(output->wc.threed.gspec),
			  &(output->wc.threed.w3D),
			  &(output->wc.threed.wspec),
			  input.quiet);
  
  if (status!=0) {
    fprintf (stderr, "Error %d reading 3D cloud file %s\n", status, wcloud3d_filename);
    return status;
  }
  
  fprintf (stderr, " ... read %d x %d x %d cloud data set\n", 
	   output->wc.threed.Nx, output->wc.threed.Ny, output->wc.threed.nlyr);

  /* we got the 3D cloud structure in output->wc.threed now; */
  /* this structure is passed to mystic()                    */
  
  
  for (kc=0; kc<output->wc.threed.nlyr; kc++)
    if (output->wc.threed.threed[kc])
      output->wc.threed.nthreed++;
  
  if (!input.quiet)
    fprintf (stderr, " ... found %d 3D layers\n", output->wc.threed.nthreed);
    
  if (!input.quiet && output->wc.threed.wspec==0)
    fprintf (stderr, " ... single scattering albedo of 3D cloud not specified - using 1D data\n");
  
  
  if (output->wc.threed.nonHGcld!=0 && output->wc.threed.gspec<2) {
    fprintf (stderr, "Error, need to specify effective droplet radius in %s.\n", wcloud3d_filename);
    fprintf (stderr, "Otherwise cannot calculate cloud phase function from the moments file!\n");
    fprintf (stderr, "gspec = %d\n", output->wc.threed.gspec);
    return -1;
  }
  
  if (output->wc.threed.nonHGcld==0 && output->wc.threed.gspec>=2) {
    fprintf (stderr, "Error, need to specify a cloud phase function as input to MYSTIC.\n");
    fprintf (stderr, "Otherwise cannot calculate cloud phase function for\n");
    fprintf (stderr, "the effective radius given in %s!\n", wcloud3d_filename);
    return -1;
  }
  
  switch (output->wc.threed.gspec) {
  case GSPEC_NONE:
    if (!input.quiet)
      fprintf (stderr, " ... asymmetry parameter of 3D cloud not specified - using 1D data\n");
    break;
  case GSPEC_EXT_GG:
    if (!input.quiet)
      fprintf (stderr, " ... reading asymmetry parameter from column 5 of the 3D cloud file\n");
    break;
  case GSPEC_EXT_REFF:
    if (!input.quiet)
      fprintf (stderr, " ... reading effective droplet radius from column 5 of the 3D cloud file\n");
    break;
  case GSPEC_LWC_REFF:
    if (!input.quiet) {
      fprintf (stderr, " ... reading effective droplet radius from column 5 of the 3D cloud file\n");
      fprintf (stderr, " ... and interpreting column 4 as liquid water content [g/m3].\n");
    }
    break;
  default:
    fprintf (stderr, "Error, gspec = %d out of range\n", output->wc.threed.gspec);
    return -1;
  }

  if (!input.quiet)
    fprintf (stderr, " ... checking 3D levels\n");
  
  /* make sure that 1D and 3D properties are defined on the same altitude grid */
  fprintf (stderr, "Attention: level consistency is not yet checked in setup_atmosphere3D()!\n");
  fprintf (stderr, "This needs to be changed!\n");
  
  /* We don't check it here because it does not work yet! We need to compare 
     with the inverted atmospheric profile which is not yet available
     at this place in the code
   status = compare_levels_df (output->wc.threed.zd, output->wc.threed.nlyr,
			      output->atm.zd_common, output->atm.nlyr);
  if (status!=0) {
    fprintf (stderr, "\nFATAL error: z-grid of 1D atmosphere and 3D clouds are different\n");
    return status;
  }
  */


  if (!input.quiet) {
    fprintf (stderr, " ... 3D cloud definition seems to be o.k.\n");
    
    for (kc=0; kc<output->wc.threed.nlyr; kc++)
      if (output->wc.threed.threed[kc])
	fprintf (stderr, " ... found 3D clouds at level %d, altitude range %g - %g\n", 
		 kc, output->wc.threed.zd[kc], output->wc.threed.zd[kc+1]);
  }

  return 0;
}


/***********************************************************************************/
/* Function: compare_levels_df                                            @62_30i@ */
/* Description:                                                                    */
/*  Compare two altitude grids. First is expected to be double, second float       */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i62_30@ */
/***********************************************************************************/

static int compare_levels_df (double *z1, int n1,
			      float *z2, int n2)
{
  int kc=0;

  if (n1!=n2) {
    fprintf (stderr, "Error, number of layers differing\n");
    return -1;
  }

  for (kc=0; kc<=n1; kc++)
    if (fabs((float) z1[kc] - z2[kc]) > MC_EPSILON * z1[kc]) {
      fprintf (stderr, "Error, difference at level %d, altitude %g vs. %g\n",	
	       kc, z1[kc], z2[kc]);
      return -1;
    }

  return 0;  /* if o.k. */
}
      




/***********************************************************************************/
/* Function: calloc_profile                                               @62_30i@ */
/* Description:                                                                    */
/*  Allocate memory for struct profile and initialize structure.                   */ 
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i62_30@ */
/***********************************************************************************/

profile *calloc_profile (int n)
{
  profile *p = calloc(1, sizeof(profile));
  if (p==NULL)
    return NULL;

  p->n = n;

  p->air = calloc((size_t) n, sizeof(double));
  if (p->air==NULL)
    return NULL;

  p->aer = calloc((size_t) n, sizeof(double));
  if (p->aer==NULL)
    return NULL;

  p->ozo = calloc((size_t) n, sizeof(double));
  if (p->ozo==NULL)
    return NULL;

  p->cld = calloc((size_t) n, sizeof(double));
  if (p->cld==NULL)
    return NULL;

  p->ice = calloc((size_t) n, sizeof(double));
  if (p->ice==NULL)
    return NULL;

  p->tot = calloc((size_t) n, sizeof(double));
  if (p->tot==NULL)
    return NULL;
  
  return p;
}


/***********************************************************************************/
/* Function: calloc_profile3D                                             @62_30i@ */
/* Description:                                                                    */
/*  Allocate memory for struct profile3D and initialize structure.                 */ 
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i62_30@ */
/***********************************************************************************/

profile3D *calloc_profile3D (int Nz, int Nx, int Ny, int *threed)
{
  int kc=0, ic=0;
  profile3D *p = calloc(1, sizeof(profile3D));
  if (p==NULL)
    return NULL;

  p->Nz = Nz;
  p->Nx = Nx;
  p->Ny = Ny;

  p->threed = calloc((size_t) Nz, sizeof(int));
  if (p->threed==NULL)
    return NULL;
  
  for (kc=0; kc<Nz; kc++) {
    p->threed[kc] = threed[kc];
    if (p->threed[kc])
      p->nthreed++;
  }


  p->cld = calloc((size_t) Nz, sizeof(float **));
  if (p->cld==NULL)
    return NULL;

  p->tot = calloc((size_t) Nz, sizeof(float **));
  if (p->tot==NULL)
    return NULL;
  
  for (kc=0; kc<Nz; kc++)
    if (p->threed[kc]) {

      p->cld[kc] = calloc((size_t) Nx, sizeof(float *));
      for (ic=0; ic<Nx; ic++)
	p->cld[kc][ic] = calloc((size_t) Ny, sizeof(float));
      
      p->tot[kc] = calloc((size_t) Nx, sizeof(float *));
      for (ic=0; ic<Nx; ic++)
	p->tot[kc][ic] = calloc((size_t) Ny, sizeof(float));
    }

  return p; 
}


/***********************************************************************************/
/* Function: setup_profiles1D                                             @62_30i@ */
/* Description:                                                                    */
/*  Initialize the 1D part of the atmosphere_struct with the 1D profiles passed    */
/*  as function arguments.                                                         */ 
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i62_30@ */
/***********************************************************************************/

int setup_profiles1D (float *dtair, float *dto3, 
		      float *dtaer, float *omaer, float *g1aer, float *g2aer, float *faer,
		      float *dtcld, float *omcld, float *g1cld, float *g2cld, float *fcld,
		      float *dtice, float *omice, float *g1ice, float *g2ice, float *fice,
		      float *lwccld, float *recld,
		      float *zprof, int nlyr,
		      atmosphere_struct *atmos)
{
  int lc=0, status=0;

  /* allocate memory for 1D profiles in atmosphere_struct *atmos */
  status = calloc_1D_atmos (atmos, nlyr);
  if (status!=0) {
    fprintf (stderr, "Error %d returned by calloc_1D_atmos()\n", status);
    return status;
  }

  /* !!! attention !!!  z[0 ... atmos->Nz], (dsca->air)[0 ... atmos->Nz-1], ...     */
  /* distinguish between layer (0..atmos->Nz-1) and level (0..atmos->Nz) properties */

  /* convert altitude km -> m      */
  for (lc=0; lc<atmos->Nz+1; lc++)
    atmos->Z[lc] = zprof[lc]*1000.0;

  /* calculate Rayleigh scattering */
  for (lc=0; lc<atmos->Nz; lc++)
    (atmos->dtsca->air)[lc] = dtair[lc];

  /* calculate ozone absorption */
  for (lc=0; lc<atmos->Nz; lc++)
    (atmos->dtabs->ozo)[lc] = dto3[lc];

  /* calculate aerosol scattering and absorption */
  for (lc=0; lc<atmos->Nz; lc++) {
    (atmos->dtsca->aer)[lc] = dtaer[lc] * omaer[lc];
    (atmos->dtabs->aer)[lc] = dtaer[lc] * (1.0 - omaer[lc]);
  }

  /* calculate cloud scattering and absorption */
  for (lc=0; lc<atmos->Nz; lc++) {
    (atmos->dtsca->cld)[lc] = dtcld[lc] * omcld[lc];
    (atmos->dtabs->cld)[lc] = dtcld[lc] * (1.0 - omcld[lc]);
  }

  /* calculate ice cloud scattering and absorption */
  for (lc=0; lc<atmos->Nz; lc++) {
    (atmos->dtsca->ice)[lc] = dtice[lc] * omice[lc];
    (atmos->dtabs->ice)[lc] = dtice[lc] * (1.0 - omice[lc]);
  }

  /* calculate total scattering optical depth */
  for (lc=0; lc<atmos->Nz; lc++) {
    (atmos->dtsca->tot)[lc] = atmos->dtsca->air[lc] + atmos->dtsca->aer[lc] + 
                              atmos->dtsca->cld[lc] + atmos->dtsca->ice[lc];

    (atmos->dtabs->tot)[lc] = atmos->dtabs->ozo[lc] + atmos->dtabs->aer[lc] + 
                              atmos->dtabs->cld[lc] + atmos->dtabs->ice[lc];

    /* scattering */
    atmos->ksca->air[lc] = atmos->dtsca->air[lc] / (atmos->Z[lc+1] - atmos->Z[lc]);
    atmos->ksca->aer[lc] = atmos->dtsca->aer[lc] / (atmos->Z[lc+1] - atmos->Z[lc]);
    atmos->ksca->cld[lc] = atmos->dtsca->cld[lc] / (atmos->Z[lc+1] - atmos->Z[lc]);
    atmos->ksca->ice[lc] = atmos->dtsca->ice[lc] / (atmos->Z[lc+1] - atmos->Z[lc]);
    atmos->ksca->tot[lc] = atmos->dtsca->tot[lc] / (atmos->Z[lc+1] - atmos->Z[lc]);

    /* absorption */
    atmos->kabs->ozo[lc] = atmos->dtabs->ozo[lc] / (atmos->Z[lc+1] - atmos->Z[lc]);
    atmos->kabs->aer[lc] = atmos->dtabs->aer[lc] / (atmos->Z[lc+1] - atmos->Z[lc]);
    atmos->kabs->cld[lc] = atmos->dtabs->cld[lc] / (atmos->Z[lc+1] - atmos->Z[lc]);
    atmos->kabs->ice[lc] = atmos->dtabs->ice[lc] / (atmos->Z[lc+1] - atmos->Z[lc]);
    atmos->kabs->tot[lc] = atmos->dtabs->tot[lc] / (atmos->Z[lc+1] - atmos->Z[lc]);
  }

  /* convert asymmetry factor to double */
  for (lc=0; lc<atmos->Nz; lc++) {
    atmos->g1->aer[lc] = g1aer[lc];
    atmos->g2->aer[lc] = g2aer[lc];
    atmos->ff->aer[lc] = faer [lc];

    atmos->g1->cld[lc] = g1cld[lc];
    atmos->g2->cld[lc] = g2cld[lc];
    atmos->ff->cld[lc] = fcld [lc];

    atmos->g1->ice[lc] = g1ice[lc];
    atmos->g2->ice[lc] = g2ice[lc];
    atmos->ff->ice[lc] = fice [lc];
  }


  for (lc=0; lc<atmos->Nz; lc++) {
    atmos->lwc->cld[lc]  = lwccld[lc];
    atmos->reff->cld[lc] = recld[lc];
  }


  /* calculate total optical depths */

  atmos->tabstot.air = 0.0;
  atmos->tabstot.aer = 0.0;
  atmos->tabstot.ozo = 0.0;
  atmos->tabstot.cld = 0.0;
  atmos->tabstot.ice = 0.0;
  atmos->tabstot.tot = 0.0;
  
  atmos->tscatot.air = 0.0;
  atmos->tscatot.aer = 0.0;
  atmos->tscatot.ozo = 0.0;
  atmos->tscatot.cld = 0.0;
  atmos->tscatot.ice = 0.0;
  atmos->tscatot.tot = 0.0;

  for (lc=0; lc<atmos->Nz; lc++) {
    atmos->tabstot.ozo += atmos->dtabs->ozo[lc]; 
    atmos->tabstot.aer += atmos->dtabs->aer[lc]; 
    atmos->tabstot.cld += atmos->dtabs->cld[lc]; 
    atmos->tabstot.ice += atmos->dtabs->ice[lc]; 
    atmos->tabstot.tot += atmos->dtabs->tot[lc];
			  
    atmos->tscatot.air += atmos->dtsca->air[lc];
    atmos->tscatot.aer += atmos->dtsca->aer[lc];
    atmos->tscatot.cld += atmos->dtsca->cld[lc];
    atmos->tscatot.ice += atmos->dtsca->ice[lc];
    atmos->tscatot.tot += atmos->dtsca->tot[lc];
  }


  return 0;
}


/***********************************************************************************/
/* Function: calloc_1D_atmos                                              @62_30i@ */
/* Description:                                                                    */
/*  Allocate memory for the 1D part of atmosphere_struct *atmos.                   */
/*                                                                                 */
/* Parameters:                                                                     */
/* Return value:                                                                   */
/* Example:                                                                        */
/* Files:                                                                          */
/* Known bugs:                                                                     */
/* Author:                                                                         */
/*                                                                        @i62_30@ */
/***********************************************************************************/

int calloc_1D_atmos (atmosphere_struct *atmos, int nlyr)
{

  atmos->Nz = nlyr;

  /* allocate memory for profiles */
  atmos->Z = calloc ((size_t) (atmos->Nz+1), sizeof(double));
  if (atmos->Z==NULL) {
    fprintf (stderr, "Error allocating memory for Z\n");
    return -1;
  }

  atmos->dtsca    = calloc_profile (atmos->Nz);
  if (atmos->dtsca==NULL) {
    fprintf (stderr, "Error allocating memory for dtsca\n");
    return -1;
  }

  atmos->dtabs    = calloc_profile (atmos->Nz);
  if (atmos->dtabs==NULL) {
    fprintf (stderr, "Error allocating memory for dtabs\n");
    return -1;
  }

  atmos->ksca     = calloc_profile (atmos->Nz);
  if (atmos->ksca==NULL)
    return -1;

  atmos->kabs     = calloc_profile (atmos->Nz);
  if (atmos->kabs==NULL) {
    fprintf (stderr, "Error allocating memory for kabs\n");
    return -1;
  }

  atmos->g1       = calloc_profile (atmos->Nz);
  if (atmos->g1==NULL) {
    fprintf (stderr, "Error allocating memory for g\n");
    return -1;
  }

  atmos->g2       = calloc_profile (atmos->Nz);
  if (atmos->g2==NULL) {
    fprintf (stderr, "Error allocating memory for g\n");
    return -1;
  }

  atmos->ff       = calloc_profile (atmos->Nz);
  if (atmos->ff==NULL) {
    fprintf (stderr, "Error allocating memory for g\n");
    return -1;
  }

  /* ??? set ff to 1 ??? */

  atmos->lwc      = calloc_profile (atmos->Nz);
  if (atmos->lwc==NULL) {
    fprintf (stderr, "Error allocating memory for lwc\n");
    return -1;
  }

  atmos->reff      = calloc_profile (atmos->Nz);
  if (atmos->reff==NULL) {
    fprintf (stderr, "Error allocating memory for reff\n");
    return -1;
  }

  return 0;
}


