/*--------------------------------------------------------------------
 * $Id: extraterrestrial.c,v 1.21 2004/10/08 06:53:35 pa2h Exp $
 * 
 * This file is part of libRadtran.
 * Copyright (c) 1997-2001 by Arve Kylling and Bernhard Mayer.
 * 
 * ######### Contact info: http://www.libradtran.org #########
 *
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License   
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.        
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of  
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the   
 * GNU General Public License for more details.                    
 * 
 * You should have received a copy of the GNU General Public License          
 * along with this program; if not, write to the Free Software                
 * Foundation, Inc., 59 Temple Place - Suite 330, 
 * Boston, MA 02111-1307, USA.
 *--------------------------------------------------------------------*/

#include <stdio.h>							    
#include <stdlib.h>
#include <string.h>

#include "uvspec.h"
#include "../src_c/ascii.h"
#include "../src_c/sun.h"

#if HAVE_LIBNETCDF
#include <netcdf.h>
#endif


/************************************/
/* prototypes of internal functions */
/************************************/

static int read_extraterrestrial(char* extraterrestrial_filename, 
				 wl_inp_struct *wl_inp, wl_out_struct *wl_out,
				 float wl_start, float wl_end, int cdf, int ck_scheme);

static int cp_lambdat_lambdah (output_struct *output);



/**************************************************************/
/* Setup extraterrestrial irradiance.                         */
/**************************************************************/

int setup_extraterrestrial (input_struct input, output_struct *output)
{
  int status=0, iv=0, cdf=0, done=0;
  char temp[FILENAME_MAX]="";
  char filename[FILENAME_MAX]="";

  float *x_flt=NULL, *y_flt=NULL;
  int n_flt=0;

  /* For special applications, pre-defined extraterrestrial */
  /* spectra are used.                                      */

  /* we don't want an extraterrestrial spectrum if */
  /* a THERMAL_BANDS_FILE is defined.              */
  if (strlen(input.filename[FN_WLBANDS])==0) {
    
    /* construct location of the extraterrestrial flux */
    strcpy (temp, input.filename[FN_PATH]);
    switch (input.ck_scheme) {
    case CK_KATO:
    case CK_KATO2:
    case CK_KATO2_96:
      strcat (temp, "solar_flux/kato");
      break;
      
    case CK_FU:
    strcat (temp, "solar_flux/fu");
    break;
    
    case CK_AVHRR_KRATZ:
      strcat (temp, "solar_flux/kratz");
      break;
      
    case CK_NONE:
    case CK_LOWTRAN:
    case CK_GENERIC:
      break;
      
    default:
      fprintf (stderr, "Error: unsupported correlated-k scheme\n");
      return -1;
    }
    
    switch (input.ck_scheme) {
    case CK_KATO:
    case CK_KATO2:
    case CK_KATO2_96:
    case CK_FU:
    case CK_AVHRR_KRATZ:
      
      /* check if we need to replace the extraterrestrial flux */
      if (!strcmp(temp, input.filename[FN_EXTRATERRESTRIAL]))
	strcpy (filename, input.filename[FN_EXTRATERRESTRIAL]);
      else 
	strcpy (filename, temp);
      
      if (!input.quiet)
	fprintf (stderr, " ... replaced extraterrestrial flux by %s\n", filename);
      
      cdf=0;
      
      break;
      
    case CK_GENERIC:
      if (!input.quiet)
	if (strlen(input.filename[FN_EXTRATERRESTRIAL]) > 0)
	  fprintf (stderr, " ... ignoring %s, reading extraterrestrial flux from %s\n",
		   input.filename[FN_EXTRATERRESTRIAL], input.filename[FN_CK_GENERIC]);

      
      strcpy (filename, input.filename[FN_CK_GENERIC]);
      cdf=1;
      break;
      
    case CK_NONE:
    case CK_LOWTRAN:
      strcpy (filename, input.filename[FN_EXTRATERRESTRIAL]);
      break;
      
    default:
      fprintf (stderr, "Error: unsupported correlated-k scheme\n");
      return -1;
    }
  }
    
  done=0;

  /* if no extraterrestrial file is defined, check if we really need it */
  if (strlen(filename)==0) {
    switch (input.calibration) {
    case OUTCAL_ABSOLUTE:
      /* in this case we really need input from a file */
      fprintf (stderr, "Error, no extraterrestrial spectrum defined!\n");
      fprintf (stderr, "Please define solar_file and retry\n");
      return -1;

      break;

    case OUTCAL_THERMAL:
    case OUTCAL_TRANSMITTANCE:
    case OUTCAL_REFLECTIVITY:
    case OUTCAL_BRIGHTNESS:

      /* if we don't have an extraterrestrial spectrum, we set it to 1 */

      if (!input.quiet)
	fprintf (stderr, " ... no extraterrestrial spectrum defined, assuming 1\n");

      status = cp_lambdat_lambdah (output);
      if (status!=0) {
	fprintf (stderr, "Error %d returned by cp_lambdat_lambdah()\n", status);
	return status;
      }

      done=1; /* no need to read the extraterrestrial file anymore */
  
      break;

    default:
      fprintf (stderr, "Error, unknown output calibration %d\n", 
	       input.calibration);
      return -1;
    }
  }

  if (!done) {
    /**** Read extraterrestrial irradiance ****/
    status = read_extraterrestrial(filename, 
				   &input.wl, &output->wl, 
				   output->wl.start, output->wl.end, cdf, input.ck_scheme);
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
  }


  if (!done) {
    /**** Allocate memory for wavelength bands ****/
    output->wl.wvnmlo_h = calloc (output->wl.nlambda_h, sizeof(float));
    output->wl.wvnmhi_h = calloc (output->wl.nlambda_h, sizeof(float));

    /**** Set to default bandwidth, not very useful ****/
    for (iv=0; iv<output->wl.nlambda_h; iv++) {
      output->wl.wvnmlo_h[iv] = 1.0E7 / output->wl.lambda_h[iv] - input.bandwidth / 2.0;
      output->wl.wvnmhi_h[iv] = output->wl.wvnmlo_h[iv] + input.bandwidth;
    }
  }

  /* allocate memory for filter function in any case */
  output->wl.filter = calloc (output->wl.nlambda_h, sizeof(float));
  
  /* set filter function to 1 (default) */
  for (iv=0; iv<output->wl.nlambda_h; iv++)
    output->wl.filter[iv] = 1;


  /**** Read filter function ****/
  if (strlen(input.filename[FN_FILTERFUNCTION])>0) {

    if (!input.quiet)
      fprintf (stderr, " ... reading filter function from  %s\n",
	       input.filename[FN_FILTERFUNCTION]);
    
    status = read_2c_file_float (input.filename[FN_FILTERFUNCTION], &x_flt, &y_flt, &n_flt); 
    if (status!=0) {
      fprintf (stderr, "Error %d reading filter function from %s\n", 
	       status, input.filename[FN_FILTERFUNCTION]);
      return status;
    }

    status = arb_wvn (n_flt, x_flt, y_flt,  
		      output->wl.nlambda_h, output->wl.lambda_h, output->wl.filter, 1, 0);

    if (status!=0) {
      fprintf (stderr, "Error %d interpolating filter function\n", status);
      return status;
    }
  }


  /**** Correction factor for Earth-Sun distance ****/
  output->eccentricity_factor = 1.0;
  
  if (input.day_of_year > 0.0) {
    output->eccentricity_factor = eccentricity (input.day_of_year);
    for (iv=0; iv<output->wl.nlambda_h; iv++)
      output->wl.fbeam[iv] *= output->eccentricity_factor;
  }

  return 0;
}





/**************************************************************/
/* Read extraterrestrial irradiance.                          */
/**************************************************************/

static int read_extraterrestrial (char *filename, 
				  wl_inp_struct *wl_inp, wl_out_struct *wl_out,
				  float wl_start, float wl_end, int cdf, int ck_scheme) 
{
  int iv=0, nlambda=0, status=0, have_wl_start=0, have_wl_end=0;
  float *tmp_lambda=NULL, *tmp_fbeam=NULL;

#if HAVE_LIBNETCDF
  double *wvnlo=NULL, *wvnhi=NULL;
  int id_wvnlo=0, id_wvnhi=0;
  int ncid=0;
  int idd_nwvn=0;
  int id_extra=0;
    
  int start[1] = {0};
  int count[1] = {0};
#endif


  /* read extraterrestrial irradiance */
  if (!cdf) {
    status = read_2c_file_float (filename, 
				 &tmp_lambda, &tmp_fbeam, &(wl_out->nlambda_h));
    
    if (status!=0) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
  }
  else {

#if HAVE_LIBNETCDF

    /* open netcdf file */
    status = nc_open (filename, NC_NOWRITE, &ncid);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d opening netCDF file %s\n", status, filename);
      return status;
    }

    /* get dimension id for "nwvn" */
    status = nc_inq_dimid (ncid, "nwvn", &idd_nwvn);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
  
    /* get dimension length for "nwvn" */
    status = nc_inq_dimlen (ncid, idd_nwvn, &(wl_out->nlambda_h));
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
    
    start[0] = 0;
    count[0] = wl_out->nlambda_h;
    
    /* allocate memory for fields */
    tmp_fbeam  = calloc (wl_out->nlambda_h, sizeof(float));
    tmp_lambda = calloc (wl_out->nlambda_h, sizeof(float));
    wvnlo      = calloc (wl_out->nlambda_h, sizeof(double));
    wvnhi      = calloc (wl_out->nlambda_h, sizeof(double));

    /* get variable id for "wvnlo" */
    status = nc_inq_varid (ncid, "wvnlo", &id_wvnlo);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
    
    /* read "wvnlo" */
    status = nc_get_vara_double (ncid, id_wvnlo, start, count, wvnlo);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }

    /* get variable id for "wvnhi" */
    status = nc_inq_varid (ncid, "wvnhi", &id_wvnhi);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }
    
    /* read "wvnhi" */
    status = nc_get_vara_double (ncid, id_wvnhi, start, count, wvnhi);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }

    for (iv=0; iv<wl_out->nlambda_h; iv++)
      tmp_lambda[iv] = 1.0E7 / (0.5 * (wvnlo[iv] + wvnhi[iv]));



    /* get variable id for "extra" */
    status = nc_inq_varid (ncid, "extra", &id_extra);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading extraterrestrial irradiance from %s\n", 
	       status, filename);
      return status;
    }
    
    /* read "extra" */
    status = nc_get_vara_float (ncid, id_extra, start, count, tmp_fbeam);
    if (status!=NC_NOERR) {
      fprintf (stderr, "Error %d reading %s\n", status, filename);
      return status;
    }

    nc_close(ncid);

#else
  fprintf (stderr, " ***********************************************************************\n");
  fprintf (stderr, " * You have built uvspec without libnetcdf and hence cannot            *\n");
  fprintf (stderr, " * use the generic correlated_k option. Please get netcdf and rebuild. *\n");
  fprintf (stderr, " ***********************************************************************\n");
  return -1;
#endif

  }

  /* count the extraterrestrial wavelengths within the */
  /* user-defined wavelength range                     */
  nlambda = 0;
  for (iv=0; iv<wl_out->nlambda_h; iv++)
    if ((tmp_lambda[iv]>=wl_start) && (tmp_lambda[iv]<=wl_end))
      nlambda++;

  wl_out->lambda_h = (float *) calloc (nlambda, sizeof(float));
  wl_out->fbeam    = (float *) calloc (nlambda, sizeof(float));

  wl_out->map_e2h  = (int *)   calloc (nlambda, sizeof(int));

  if (ck_scheme == CK_NONE) {

    /* Check if the extraterrestrial spectrum covers */
    /* the required wavelength range                 */
    if (wl_start < tmp_lambda[0]) {
      fprintf (stderr, "Error: required start wavelength: %.3f nm outside extraterrestrial\n",
	       wl_start);
      fprintf (stderr, "       spectrum range [%.3f - %.3f]\n", 
	       tmp_lambda[0], tmp_lambda[wl_out->nlambda_h-1]);
      return -1;
    }

    if ( wl_end > tmp_lambda[wl_out->nlambda_h-1] ) {
      fprintf (stderr, "Error: required end wavelength: %.3f nm outside extraterrestrial\n", 
	       wl_end);
      fprintf (stderr, "       spectrum range [%.3f - %.3f]\n", 
	       tmp_lambda[0], tmp_lambda[wl_out->nlambda_h-1]);
      return -1;
    }

    /* Both wl_end and wl_start should be in the extraterrestrial solar file         */
    /* Otherwise problems may occurr below if a single wavelength which is not       */
    /* part of the solar flux file is specified with the wvn option. The user should */
    /* use the spline option to get arbitrary wavelengths or include the wanted      */
    /* wavelength in the extraterrestrial solar flux file.                           */
    for (iv=0; iv<wl_out->nlambda_h; iv++) {
      if (wl_start == tmp_lambda[iv] ) have_wl_start=1;
      if (wl_end   == tmp_lambda[iv] ) have_wl_end=1;
    }
    if (have_wl_start==0) {
      fprintf (stderr, "Error: required start wavelength: %.6f nm is not in extraterrestrial solar flux file.\n", 
	       wl_start);
      fprintf (stderr, "       Either include wavelength in solar flux file, or\n");
      fprintf (stderr, "       or use spline option.\n");      
      return -1;
    }
    if (have_wl_end==0) {
      fprintf (stderr, "Error: required ened wavelength: %.6f nm is not in extraterrestrial solar flux file.\n", 
	       wl_end);
      fprintf (stderr, "       Either include wavelength in solar flux file, or\n");
      fprintf (stderr, "       or use spline option.\n");      
      return -1;
    }
  }


  nlambda = 0;

  for (iv=0; iv<wl_out->nlambda_h; iv++)
    if ((tmp_lambda[iv]>=wl_start) && (tmp_lambda[iv]<=wl_end)) {
      wl_out->lambda_h[nlambda] = tmp_lambda[iv];
      wl_out->fbeam[nlambda]    = tmp_fbeam[iv];
      wl_out->map_e2h[nlambda]  = iv;
      nlambda++;
    }

  wl_out->nlambda_h = nlambda;

  /* free memory */
  free(tmp_lambda);
  free(tmp_fbeam);

  return 0;
}



/* copy the internal (computational) wavelength grid */
/* to the output grid                                */
static int cp_lambdat_lambdah (output_struct *output) 
{
  int iv=0, nlambda=0;

  output->wl.nlambda_h = output->wl.nlambda_t;
  output->wl.lambda_h  = calloc (output->wl.nlambda_h, sizeof(float));
  output->wl.fbeam     = calloc (output->wl.nlambda_h, sizeof(float));
  output->wl.map_e2h   = calloc (output->wl.nlambda_h, sizeof(int));

  output->wl.wvnmlo_h = calloc (output->wl.nlambda_h, sizeof(float));
  output->wl.wvnmhi_h = calloc (output->wl.nlambda_h, sizeof(float));
  
  for (iv=0; iv<output->wl.nlambda_h; iv++)
    if ((output->wl.lambda_t[iv]>=output->wl.start) && (output->wl.lambda_t[iv]<=output->wl.end)) {
      output->wl.lambda_h[nlambda] = output->wl.lambda_t[iv];
      output->wl.wvnmlo_h[nlambda] = output->wl.wvnmlo_t[iv];
      output->wl.wvnmhi_h[nlambda] = output->wl.wvnmhi_t[iv];
      output->wl.fbeam[nlambda]    = 1.0;
      output->wl.map_e2h[nlambda]  = iv;
      nlambda++;
    }
  
  output->wl.nlambda_h = nlambda;
  
  return 0;
}
